/* rangeGraph - test ideas about graph that tracks possible distance ranges. */
#include "common.h"
#include "linefile.h"
#include "hash.h"
#include "diGraph.h"
#include "rgi.h"

void usage()
/* Explain usage and exit. */
{
errAbort(
  "rangeGraph - test ideas about graph that tracks possible distance ranges\n"
  "usage:\n"
  "   rangeGraph XXX\n");
}

struct rgi *readRgi(char *inName)
{
struct rgi *rgiList = NULL, *rgi;
struct lineFile *lf = lineFileOpen(inName, TRUE);
int wordCount;
char *words[8];

while ((wordCount = lineFileChop(lf, words)) != 0)
    {
    lineFileExpectWords(lf, 4, wordCount);
    rgi = rgiLoad(words);
    slAddHead(&rgiList, rgi);
    uglyf("%s %s: min %d, max %d\n", rgi->a, rgi->b, rgi->minDistance, rgi->maxDistance);
    }
lineFileClose(&lf);
slReverse(&rgiList);
return rgiList;
}

struct edgeRange
/* A distance range. */
   {
   int minDist;	/* Must be at least this long. */
   int maxDist;     /* But no longer than this. */
   };

struct nodeRange
/* A distance range. */
   {
   int minDist;	/* Must be at least this long. */
   int maxDist;     /* But no longer than this. */
   };

void maybeMakeNode(struct diGraph *graph, char *nodeName)
/* Make node in graph if not there already. */
{
if (!dgFindNode(graph, nodeName))
    {
    struct nodeRange *r;
    AllocVar(r);
    dgAddNode(graph, nodeName, r);
    }
}

boolean parentsAllProcessed(struct dgNode *node)
/* Return TRUE if all parents of node have been processed. */
{
struct dgConnection *conn;
struct dgNode *parent;

for (conn = node->prevList; conn != NULL; conn = conn->next)
    {
    if (!conn->node->conn)
        return FALSE;
    }
return TRUE;
}

void addToPending(struct dgNode *graphNode, struct dlList *pendingList, struct dlList *freeList)
/* Add graph node to pending list, if possibly recycling a list node from freeList to 
 * do it. */
{
struct dlNode *listNode;
if ((listNode = dlPopHead(freeList)) != NULL)
    {
    listNode->val = graphNode;
    dlAddTail(pendingList, listNode);
    }
else
    {
    dlAddValTail(pendingList, graphNode);
    }
}

boolean testOneStart(struct diGraph *graph, struct dgNode *rootNode, struct dlList *pendingList,
    struct dlList *freeList)
/* Fanning out from rootNode, make sure graph looks consistent. */
{
boolean allOk = TRUE;
struct dlNode *listNode;
struct dgNode *graphNode, *child;
struct dgEdge *edge;
struct dgConnection *wayOut;
boolean gotOne;
struct edgeRange *edgeRange;
struct nodeRange *nodeRange, *childRange;

/* Initialize pending list to only have root node on it. */
dlCat(freeList, pendingList);
addToPending(rootNode, pendingList, freeList);

/* We'll use visit flags to tell if this is the first time a
 * range has been assigned to a vertex.  First need to clear them. */
dgClearVisitFlags(graph);

/* Loop through pending list looking for nodes we can process and then
 * processing them until the list pending list is empty. */
while (allOk)
    {
    gotOne = FALSE;
    for (listNode = pendingList->head; listNode->next != NULL; listNode = listNode->next)
        {
	graphNode = listNode->val;
	if (parentsAllProcessed(graphNode))
	    {
	    gotOne = TRUE;
	    break;
	    }
	}

    /* If we didn't get one with no parents we should be done. */
    if (!gotOne)
        {
	break;		/* Breaks out and leads to positive return! */
	}

    /* Remove node from pending list and put it on free list for reuse. 
     * Mark it as processed. Find it's range*/
    dlRemove(listNode);
    dlAddTail(freeList, listNode);
    graphNode->conn = TRUE;
    nodeRange = graphNode->val;

    /* Look at all of the ways out of this node.  Make sure distance ranges
     * to them is consistent with our distance range. */
    for (wayOut = graphNode->nextList; wayOut != NULL; wayOut = wayOut->next)
        {
	int rMin, rMax;
	child = wayOut->node;
	childRange = child->val;
	edge = wayOut->edgeOnList->val;
	edgeRange = edge->val;
	rMin = nodeRange->minDist + edgeRange->minDist;
	rMax = nodeRange->maxDist + edgeRange->maxDist;
	if (child->visited)
	    {
	    if (childRange->minDist < rMin)
	        childRange->minDist = rMin;
	    if (childRange->maxDist > rMax)
	        childRange->maxDist = rMax;
	    if (childRange->minDist > childRange->maxDist)
	        {
		allOk = FALSE;
		break;
		}
	    }
	else
	    {
	    childRange->minDist = rMin;
	    childRange->maxDist = rMax;
	    child->visited = TRUE;
	    }
	/* Put child on pending list. */
	addToPending(child, pendingList, freeList);
	}
    }
return allOk;
}

void clearNodeRangesAndConnFlag(struct diGraph *graph)
/* Clear all values of node ranges. */
{
struct dgNode *graphNode;
struct nodeRange *nodeRange, *childRange;
for (graphNode = graph->nodeList; graphNode != NULL; graphNode = graphNode->next)
    {
    graphNode->conn = FALSE;
    nodeRange = graphNode->val;
    nodeRange->minDist = nodeRange->maxDist = 0;
    }
}

void dgClearFlagA(struct diGraph *graph)
/* Clear flagA's  in all nodes. */
{
struct dgNode *graphNode;
for (graphNode = graph->nodeList; graphNode != NULL; graphNode = graphNode->next)
    graphNode->flagA = FALSE;
}

boolean rangeGraphConsistent(struct diGraph *graph)
/* Return TRUE if range graph is internally consistent. */
{
struct dlList *pendingList = newDlList();	/* Vertices to examine next. */
struct dlList *freeList = newDlList();          /* dlNodes we can reuse. */
struct dlList *startList = newDlList();         /* Starting dlNodes. */
struct dlList *newStartList = newDlList();      /* Start nodes for next time. */
struct dlList *finishedList = newDlList();     /* Node we don't need to process. */
struct dlList *swapper;
boolean allOk = TRUE;
struct dgNode *graphNode, *child;
struct dlNode *listNode;
struct dgConnection *conn;

if (dgHasCycles(graph))
    return FALSE;

/* Use flagA to show that we've started from this node already. */
dgClearFlagA(graph);

/* Add all nodes to list that lack parents but do have children. */
for (graphNode = graph->nodeList; graphNode != NULL; graphNode = graphNode->next)
    {
    /* Initial starting nodes are those with no ancestors. */
    if (graphNode->prevList == NULL)
        {
	if (graphNode->nextList != NULL)
	    {
	    dlAddValTail(startList, graphNode);
	    }
	else
	    {
	    /* Remove isolated nodes from further consideration. */
	    dlAddValTail(finishedList, graphNode);
	    }
	graphNode->flagA = TRUE;
	}
    }

while (!dlEmpty(startList))
    {
#ifdef DEBUG
        {
	uglyf("  start list");
	for (listNode = startList->head; !dlEnd(listNode); listNode = listNode->next)
	    {
	    graphNode  = listNode->val;
	    uglyf(" %s", graphNode->name);
	    }
	uglyf(", finished list");
	for (listNode = finishedList->head; !dlEnd(listNode); listNode = listNode->next)
	    {
	    graphNode  = listNode->val;
	    uglyf(" %s", graphNode->name);
	    }
	uglyf("\n");
	}
#endif /* DEBUG */

    /* We'll use conn flags to signal that a node has been completely
     * processed (for the all predecessors processed check.) */
    clearNodeRangesAndConnFlag(graph);

    for (listNode = finishedList->head; !dlEnd(listNode); listNode = listNode->next)
        {
	graphNode = listNode->val;
	graphNode->conn = TRUE;
	}

    /* Test from each starting node. */
    for (listNode = startList->head; !dlEnd(listNode); listNode = listNode->next)
        {
	graphNode = listNode->val;
	if (!testOneStart(graph, graphNode, pendingList, freeList))
	    {
	    allOk = FALSE;
	    break;
	    }
	}
    if (!allOk)
        break;

    /* For my peace of mind check that all nodes actually did get processed. */
    if (allOk)
	{
	for (graphNode = graph->nodeList; graphNode != NULL; graphNode = graphNode->next)
	    if (!graphNode->conn)
		warn("Node %s never processed!", graphNode->name);
	}

    /* Move start list to finished list.  Add non-previously processed
     * children of start list to new start list. */
    while ((listNode = dlPopHead(startList)) != NULL)
        {
	dlAddTail(finishedList, listNode);
	graphNode = listNode->val;
	for (conn = graphNode->nextList; conn != NULL; conn = conn->next)
	    {
	    child = conn->node;
	    if (!child->flagA)
	        {
		child->flagA = TRUE;
		dlAddValTail(newStartList, child);
		}
	    }
	}
    /* Swap now empty startList and newStartList. */
    swapper = newStartList;
    newStartList = startList;
    startList = swapper;       
    }

freeDlList(&startList);
freeDlList(&newStartList);
freeDlList(&finishedList);
freeDlList(&pendingList);
freeDlList(&freeList);
return allOk;
}

boolean tryAddingRangeEdge(struct diGraph *graph, struct dgNode *a, struct dgNode *b,
	int minDist, int maxDist)
/* Try adding node from a to b to graph with the given min and max distance
 * constraints.  Return FALSE if can't. */
{
struct edgeRange oldRange, *r;
boolean gotOldRange = FALSE;
struct dgEdge *edge;

edge = dgDirectlyFollows(graph, a, b);
if (edge != NULL)
    {
    /* Save old value of edge in case have to back out. */
    r = edge->val;
    oldRange = *r;
    gotOldRange = TRUE;
    }
else
   {
   edge = dgConnect(graph, a, b);
   AllocVar(r);
   edge->val = r;
   }
r->minDist = minDist;
r->maxDist = maxDist;
if (!rangeGraphConsistent(graph))
    {
    if (gotOldRange)
        {
	*r = oldRange;
	}
    else
        {
	freeMem(&edge->val);
	dgDisconnect(graph, a, b);
	}
    return FALSE;
    }
else
    return TRUE;
}

void rangeGraph(char *inName)
/* rangeGraph - test ideas about graph that tracks possible distance ranges. */
{
struct rgi *rgiList, *rgi;
struct diGraph *graph = dgNew();
boolean ok;

/* Read in range list and add all nodes to graph. */
rgiList = readRgi(inName);
for (rgi = rgiList; rgi != NULL; rgi = rgi->next)
    {
    maybeMakeNode(graph, rgi->a);
    maybeMakeNode(graph, rgi->b);
    }

for (rgi = rgiList; rgi != NULL; rgi = rgi->next)
    {
    struct dgNode *a, *b;
    a = dgFindNode(graph, rgi->a);
    b = dgFindNode(graph, rgi->b);
    ok = tryAddingRangeEdge(graph, a, b, rgi->minDistance, rgi->maxDistance);
    printf("Added %s->%s min %d, max %d. %s\n",
    	rgi->a, rgi->b, rgi->minDistance, rgi->maxDistance,
	(ok ? "ok" : "conflict") );
    }
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc != 2)
    usage();
rangeGraph(argv[1]);
return 0;
}
